@@ -1,15 +1,18 @@
 import abc
+import inspect
+import numbers
 import textwrap
-import warnings
 from copy import deepcopy
-from typing import Any, Tuple, Union, Iterable, Optional
+from typing import Any
 from collections import namedtuple
-from collections.abc import Mapping
+from collections.abc import Mapping, Iterable
+
+import numpy as np
 
 import astropy.nddata
 import astropy.units as u
-import numpy as np
 from astropy.units import UnitsError
+from astropy.wcs.utils import _split_matrix
 
 try:
     # Import sunpy coordinates if available to register the frames and WCS functions with astropy
@@ -18,15 +21,16 @@ except ImportError:
     pass
 
 from astropy.wcs import WCS
-from astropy.wcs.utils import _split_matrix
 from astropy.wcs.wcsapi import BaseHighLevelWCS, HighLevelWCSWrapper
+from astropy.wcs.wcsapi.high_level_api import values_to_high_level_objects
 
 from ndcube import utils
 from ndcube.extra_coords.extra_coords import ExtraCoords, ExtraCoordsABC
 from ndcube.global_coords import GlobalCoords, GlobalCoordsABC
+from ndcube.meta import NDMetaABC
 from ndcube.mixins import NDCubeSlicingMixin
 from ndcube.ndcube_sequence import NDCubeSequence
-from ndcube.utils.wcs_high_level_conversion import values_to_high_level_objects
+from ndcube.utils.exceptions import warn_deprecated, warn_user
 from ndcube.visualization import PlotterDescriptor
 from ndcube.wcs.wrappers import CompoundLowLevelWCS, ResampledLowLevelWCS
 
@@ -73,7 +77,7 @@ class NDCubeABC(astropy.nddata.NDDataBase):
 
     @property
     @abc.abstractmethod
-    def array_axis_physical_types(self) -> Iterable[Tuple[str, ...]]:
+    def array_axis_physical_types(self) -> Iterable[tuple[str, ...]]:
         """
         Returns the WCS physical types that vary along each array axis.
 
@@ -90,9 +94,9 @@ class NDCubeABC(astropy.nddata.NDDataBase):
 
     @abc.abstractmethod
     def axis_world_coords(self,
-                          *axes: Union[int, str],
+                          *axes: int | str,
                           pixel_corners: bool = False,
-                          wcs: Optional[Union[BaseHighLevelWCS, ExtraCoordsABC]] = None
+                          wcs: BaseHighLevelWCS | ExtraCoordsABC | None = None
                           ) -> Iterable[Any]:
         """
         Returns objects representing the world coordinates of pixel centers for a desired axes.
@@ -134,9 +138,9 @@ class NDCubeABC(astropy.nddata.NDDataBase):
 
     @abc.abstractmethod
     def axis_world_coords_values(self,
-                                 *axes: Union[int, str],
+                                 *axes: int | str,
                                  pixel_corners: bool = False,
-                                 wcs: Optional[Union[BaseHighLevelWCS, ExtraCoordsABC]] = None
+                                 wcs: BaseHighLevelWCS | ExtraCoordsABC | None = None
                                  ) -> Iterable[u.Quantity]:
         """
         Returns the world coordinate values of all pixels for desired axes.
@@ -183,7 +187,8 @@ class NDCubeABC(astropy.nddata.NDDataBase):
     @abc.abstractmethod
     def crop(self,
              *points: Iterable[Any],
-             wcs: Optional[Union[BaseHighLevelWCS, ExtraCoordsABC]] = None
+             wcs: BaseHighLevelWCS | ExtraCoordsABC | None = None,
+             keepdims: bool = False,
              ) -> "NDCubeABC":
         """
         Crop using real world coordinates.
@@ -212,6 +217,10 @@ class NDCubeABC(astropy.nddata.NDDataBase):
             could be used it is expected that either the ``.wcs`` or
             ``.extra_coords`` properties will be used.
 
+        keepdims: `bool`, optional
+            If `False` and if cropping results in length-1 dimensions, these are sliced away in output cube.
+            If `True`, length-1 dimensions are kept.  Default=False
+
         Returns
         -------
         `~ndcube.ndcube.NDCubeABC`
@@ -226,9 +235,10 @@ class NDCubeABC(astropy.nddata.NDDataBase):
 
     @abc.abstractmethod
     def crop_by_values(self,
-                       *points: Iterable[Union[u.Quantity, float]],
-                       units: Optional[Iterable[Union[str, u.Unit]]] = None,
-                       wcs: Optional[Union[BaseHighLevelWCS, ExtraCoordsABC]] = None
+                       *points: Iterable[u.Quantity | float],
+                       units: Iterable[str | u.Unit] | None = None,
+                       wcs: BaseHighLevelWCS | ExtraCoordsABC | None = None,
+                       keepdims: bool = False
                        ) -> "NDCubeABC":
         """
         Crop using real world coordinates.
@@ -261,6 +271,10 @@ class NDCubeABC(astropy.nddata.NDDataBase):
             could be used it is expected that either the ``.wcs`` or
             ``.extra_coords`` properties will be used.
 
+        keepdims: `bool`, optional
+            If `False` and if cropping results in length-1 dimensions, these are sliced away in output cube.
+            If `True`, length-1 dimensions are kept.  Default=False
+
         Returns
         -------
         `~ndcube.ndcube.NDCubeABC`
@@ -296,7 +310,7 @@ class NDCubeLinkedDescriptor:
 
     def __get__(self, obj, objtype=None):
         if obj is None:
-            return
+            return None
 
         if getattr(obj, self._attribute_name, None) is None and self._default_type is not None:
             self.__set__(obj, self._default_type)
@@ -365,12 +379,8 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
     def __init__(self, data, wcs=None, uncertainty=None, mask=None, meta=None,
                  unit=None, copy=False, **kwargs):
 
-        super().__init__(data, uncertainty=uncertainty, mask=mask,
+        super().__init__(data, wcs=wcs, uncertainty=uncertainty, mask=mask,
                          meta=meta, unit=unit, copy=copy, **kwargs)
-        if not self.wcs:
-            self.wcs = wcs  # This line is required as a patch for an astropy bug.
-        # Above line is in if statement to prevent WCS being overwritten with None
-        # if we are instantiating from an NDCube.
 
         # Enforce that the WCS object is not None
         if self.wcs is None:
@@ -390,6 +400,46 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
                 global_coords = deepcopy(global_coords)
             self._global_coords = global_coords
 
+        # If meta is axis-aware, make it to have same shape as cube.
+        if isinstance(self.meta, NDMetaABC):
+            self.meta.data_shape = self.shape
+
+    @property
+    def data(self):
+        """
+        `~numpy.ndarray`-like : The stored dataset.
+
+        Notes
+        -----
+        It is possible to set the ``.data`` attribute on a `NDCube` with an
+        array-like object of the same shape. However, this is really only
+        intended for replacing the data with a different object representing
+        the same physical data as no other properties of the cube will be
+        changed, such as uncertainty or unit.
+        """
+        return super().data
+
+    @data.setter
+    def data(self, value):
+        # In an array-agnostic way check the shape is the same
+        if not hasattr(value, "shape") or value.shape != self.data.shape:
+            raise TypeError(f"Can only set data with an array-like object of the same shape ({self.data.shape})")
+
+        # Other masked arrays are hard to detect reliably
+        if isinstance(value, np.ma.MaskedArray):
+            raise TypeError("Can not set the .data attribute with a numpy masked array, please set .data and .mask separately.")
+
+        if isinstance(value, u.Quantity):
+            unit_error = f"Unable to set data with unit {value.unit} as it incompatible with the current unit of {self.unit}"
+            if self.unit is None:
+                raise u.UnitsError(unit_error)
+            try:
+                value = value.to_value(self.unit)
+            except u.UnitsError as exc:
+                raise u.UnitsError(unit_error) from exc
+
+        self._data = value
+
     @property
     def extra_coords(self):
         # Docstring in NDCubeABC.
@@ -411,9 +461,14 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
             CompoundLowLevelWCS(self.wcs.low_level_wcs, self._extra_coords.wcs, mapping=mapping)
         )
 
+    @property
+    def shape(self):
+        return self.data.shape
+
     @property
     def dimensions(self):
-        return u.Quantity(self.data.shape, unit=u.pix, copy=_NUMPY_COPY_IF_NEEDED)
+        warn_deprecated("Replaced by ndcube.NDCube.shape")
+        return u.Quantity(self.data.shape, unit=u.pix)
 
     @property
     def array_axis_physical_types(self):
@@ -424,24 +479,43 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         return [tuple(world_axis_physical_types[axis_correlation_matrix[:, i]].tolist())
                 for i in range(axis_correlation_matrix.shape[1])][::-1]
 
-    def _generate_world_coords(self, pixel_corners, wcs, needed_axes=None, *, units):
-        # Create meshgrid of all pixel coordinates.
-        # If user wants pixel_corners, set pixel values to pixel corners.
-        # Else make pixel centers.
+    @property
+    def quantity(self):
+        """Unitful representation of the NDCube data."""
+        return u.Quantity(self.data, self.unit, copy=_NUMPY_COPY_IF_NEEDED)
+
+    def _generate_world_coords(self, pixel_corners, wcs, *, needed_axes, units=None):
+        """
+        Private method to generate world coordinates.
+
+        Parameters
+        ----------
+        pixel_corners : bool
+            If one needs pixel corners, otherwise pixel centers.
+        wcs : astropy.wcs.WCS
+            The WCS.
+        needed_axes : array-like
+            The axes that are needed.
+        units : bool
+            If units are needed.
+
+        Returns
+        -------
+        array-like
+            The world coordinates.
+        """
         pixel_shape = self.data.shape[::-1]
         if pixel_corners:
             pixel_shape = tuple(np.array(pixel_shape) + 1)
             ranges = [np.arange(i) - 0.5 for i in pixel_shape]
         else:
             ranges = [np.arange(i) for i in pixel_shape]
-
         # Limit the pixel dimensions to the ones present in the ExtraCoords
         if isinstance(wcs, ExtraCoords):
             ranges = [ranges[i] for i in wcs.mapping]
             wcs = wcs.wcs
             if wcs is None:
-                return []
-
+                return ()
         # This value of zero will be returned as a throwaway for unneeded axes, and a numerical value is
         # required so values_to_high_level_objects in the calling function doesn't crash or warn
         world_coords = [0] * wcs.world_n_dim
@@ -455,10 +529,16 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
             # First construct a range of pixel indices for this set of coupled dimensions
             sub_range = [ranges[idx] for idx in pixel_axes_indices]
             # Then get a set of non correlated dimensions
-            non_corr_axes = set(list(range(wcs.pixel_n_dim))) - set(pixel_axes_indices)
+            non_corr_axes = set(range(wcs.pixel_n_dim)) - set(pixel_axes_indices)
             # And inject 0s for those coordinates
             for idx in non_corr_axes:
                 sub_range.insert(idx, 0)
+            # If we are subsetting world axes, ignore any pixel axes which are not correlated with our requested world axis.
+            if any(world_axis in needed_axes for world_axis in world_axes_indices):
+                needed_pixel_axes = wcs.axis_correlation_matrix[needed_axes]
+                unneeded_pixel_axes = np.argwhere(needed_pixel_axes.sum(axis=0) == 0)[:, 0]
+                for idx in unneeded_pixel_axes:
+                    sub_range[idx] = 0
             # Generate a grid of broadcastable pixel indices for all pixel dimensions
             grid = np.meshgrid(*sub_range, indexing='ij')
             # Convert to world coordinates
@@ -473,11 +553,9 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
                 array_slice[wcs.axis_correlation_matrix[idx]] = slice(None)
                 tmp_world = world[idx][tuple(array_slice)].T
                 world_coords[idx] = tmp_world
-
         if units:
             for i, (coord, unit) in enumerate(zip(world_coords, wcs.world_axis_units)):
                 world_coords[i] = coord << u.Unit(unit)
-
         return world_coords
 
     @utils.cube.sanitize_wcs
@@ -485,35 +563,27 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         # Docstring in NDCubeABC.
         if isinstance(wcs, BaseHighLevelWCS):
             wcs = wcs.low_level_wcs
-
         orig_wcs = wcs
         if isinstance(wcs, ExtraCoords):
             wcs = wcs.wcs
             if not wcs:
-                return tuple()
-
+                return ()
         object_names = np.array([wao_comp[0] for wao_comp in wcs.world_axis_object_components])
         unique_obj_names = utils.misc.unique_sorted(object_names)
         world_axes_for_obj = [np.where(object_names == name)[0] for name in unique_obj_names]
-
         # Create a mapping from world index in the WCS to object index in axes_coords
         world_index_to_object_index = {}
         for object_index, world_axes in enumerate(world_axes_for_obj):
             for world_index in world_axes:
                 world_index_to_object_index[world_index] = object_index
-
         world_indices = utils.wcs.calculate_world_indices_from_axes(wcs, axes)
         object_indices = utils.misc.unique_sorted(
             [world_index_to_object_index[world_index] for world_index in world_indices]
         )
-
-        axes_coords = self._generate_world_coords(pixel_corners, orig_wcs, world_indices, units=False)
-
+        axes_coords = self._generate_world_coords(pixel_corners, orig_wcs, needed_axes=world_indices, units=False)
         axes_coords = values_to_high_level_objects(*axes_coords, low_level_wcs=wcs)
-
         if not axes:
             return tuple(axes_coords)
-
         return tuple(axes_coords[i] for i in object_indices)
 
     @utils.cube.sanitize_wcs
@@ -521,23 +591,19 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         # Docstring in NDCubeABC.
         if isinstance(wcs, BaseHighLevelWCS):
             wcs = wcs.low_level_wcs
-
         orig_wcs = wcs
         if isinstance(wcs, ExtraCoords):
             wcs = wcs.wcs
-
+            if not wcs:
+                return ()
         world_indices = utils.wcs.calculate_world_indices_from_axes(wcs, axes)
-
-        axes_coords = self._generate_world_coords(pixel_corners, orig_wcs, world_indices, units=True)
-
+        axes_coords = self._generate_world_coords(pixel_corners, orig_wcs, needed_axes=world_indices, units=True)
         world_axis_physical_types = wcs.world_axis_physical_types
-
         # If user has supplied axes, extract only the
         # world coords that correspond to those axes.
         if axes:
             axes_coords = [axes_coords[i] for i in world_indices]
             world_axis_physical_types = tuple(np.array(world_axis_physical_types)[world_indices])
-
         # Return in array order.
         # First replace characters in physical types forbidden for namedtuple identifiers.
         identifiers = []
@@ -549,46 +615,45 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         CoordValues = namedtuple("CoordValues", identifiers)
         return CoordValues(*axes_coords[::-1])
 
-    def crop(self, *points, wcs=None):
+    def crop(self, *points, wcs=None, keepdims=False):
         # The docstring is defined in NDCubeABC
         # Calculate the array slice item corresponding to bounding box and return sliced cube.
-        item = self._get_crop_item(*points, wcs=wcs)
+        item = self._get_crop_item(*points, wcs=wcs, keepdims=keepdims)
         return self[item]
 
     @utils.cube.sanitize_wcs
-    def _get_crop_item(self, *points, wcs=None):
+    def _get_crop_item(self, *points, wcs=None, keepdims=False):
         # Sanitize inputs.
         no_op, points, wcs = utils.cube.sanitize_crop_inputs(points, wcs)
         # Quit out early if we are no-op
         if no_op:
             return tuple([slice(None)] * wcs.pixel_n_dim)
-        else:
-            comp = [c[0] for c in wcs.world_axis_object_components]
-            # Trim to unique component names - `np.unique(..., return_index=True)
-            # keeps sorting alphabetically, set() seems just nondeterministic.
-            for k, c in enumerate(comp):
-                if comp.count(c) > 1:
-                    comp.pop(k)
-            classes = [wcs.world_axis_object_classes[c][0] for c in comp]
-            for i, point in enumerate(points):
-                if len(point) != len(comp):
-                    raise ValueError(f"{len(point)} components in point {i} do not match "
-                                     f"WCS with {len(comp)} components.")
-                for j, value in enumerate(point):
-                    if not (value is None or isinstance(value, classes[j])):
-                        raise TypeError(f"{type(value)} of component {j} in point {i} is "
-                                        f"incompatible with WCS component {comp[j]} "
-                                        f"{classes[j]}.")
-            return utils.cube.get_crop_item_from_points(points, wcs, False)
+        comp = [c[0] for c in wcs.world_axis_object_components]
+        # Trim to unique component names - `np.unique(..., return_index=True)
+        # keeps sorting alphabetically, set() seems just nondeterministic.
+        for k, c in enumerate(comp):
+            if comp.count(c) > 1:
+                comp.pop(k)
+        classes = [wcs.world_axis_object_classes[c][0] for c in comp]
+        for i, point in enumerate(points):
+            if len(point) != len(comp):
+                raise ValueError(f"{len(point)} components in point {i} do not match "
+                                 f"WCS with {len(comp)} components.")
+            for j, value in enumerate(point):
+                if not (value is None or isinstance(value, classes[j])):
+                    raise TypeError(f"{type(value)} of component {j} in point {i} is "
+                                    f"incompatible with WCS component {comp[j]} "
+                                    f"{classes[j]}.")
+        return utils.cube.get_crop_item_from_points(points, wcs, False, keepdims=keepdims)
 
-    def crop_by_values(self, *points, units=None, wcs=None):
+    def crop_by_values(self, *points, units=None, wcs=None, keepdims=False):
         # The docstring is defined in NDCubeABC
         # Calculate the array slice item corresponding to bounding box and return sliced cube.
-        item = self._get_crop_by_values_item(*points, units=units, wcs=wcs)
+        item = self._get_crop_by_values_item(*points, units=units, wcs=wcs, keepdims=keepdims)
         return self[item]
 
     @utils.cube.sanitize_wcs
-    def _get_crop_by_values_item(self, *points, units=None, wcs=None):
+    def _get_crop_by_values_item(self, *points, units=None, wcs=None, keepdims=False):
         # Sanitize inputs.
         no_op, points, wcs = utils.cube.sanitize_crop_inputs(points, wcs)
         # Quit out early if we are no-op
@@ -621,19 +686,19 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
                         raise UnitsError(f"Unit '{points[i][j].unit}' of coordinate object {j} in point {i} is "
                                          f"incompatible with WCS unit '{wcs.world_axis_units[j]}'") from err
 
-        return utils.cube.get_crop_item_from_points(points, wcs, True)
+        return utils.cube.get_crop_item_from_points(points, wcs, True, keepdims=keepdims)
 
     def __str__(self):
         return textwrap.dedent(f"""\
                 NDCube
                 ------
-                Dimensions: {self.dimensions}
+                Shape: {self.shape}
                 Physical Types of Axes: {self.array_axis_physical_types}
                 Unit: {self.unit}
                 Data Type: {self.data.dtype}""")
 
     def __repr__(self):
-        return f"{object.__repr__(self)}\n{str(self)}"
+        return f"{object.__repr__(self)}\n{self!s}"
 
     def explode_along_axis(self, axis):
         """
@@ -648,9 +713,9 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         -------
         result : `ndcube.NDCubeSequence`
         """
-        # If axis is -ve then calculate the axis from the length of the dimensions of one cube
+        # If axis is -ve then calculate the axis from the length of the shape of one cube
         if axis < 0:
-            axis = len(self.dimensions) + axis
+            axis = len(self.shape) + axis
         # To store the resultant cube
         result_cubes = []
         # All slices are initially initialised as slice(None, None, None)
@@ -668,7 +733,12 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         # Creating a new NDCubeSequence with the result_cubes and common axis as axis
         return NDCubeSequence(result_cubes, meta=self.meta)
 
-    def reproject_to(self, target_wcs, algorithm='interpolation', shape_out=None, return_footprint=False, **reproject_args):
+    def reproject_to(self,
+                     target_wcs,
+                     algorithm='interpolation',
+                     shape_out=None,
+                     return_footprint=False,
+                     **reproject_args):
         """
         Reprojects the instance to the coordinates described by another WCS object.
 
@@ -702,7 +772,8 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
         Returns
         -------
         reprojected_cube : `ndcube.NDCube`
-            A new resultant NDCube object, the supplied ``target_wcs`` will be the ``.wcs`` attribute of the output `~ndcube.NDCube`.
+            A new resultant NDCube object, the supplied ``target_wcs`` will be
+            the ``.wcs`` attribute of the output `~ndcube.NDCube`.
 
         footprint: `numpy.ndarray`
             Footprint of the input array in the output array.
@@ -725,7 +796,8 @@ class NDCubeBase(NDCubeABC, astropy.nddata.NDData, NDCubeSlicingMixin):
             from reproject import reproject_adaptive, reproject_exact, reproject_interp
             from reproject.wcs_utils import has_celestial
         except ModuleNotFoundError:
-            raise ImportError("The NDCube.reproject_to method requires the optional package `reproject`.")
+            raise ImportError(f"The {type(self).__name__}.reproject_to method requires "
+                              f"the `reproject` library to be installed.")
 
         algorithms = {
             "interpolation": reproject_interp,
@@ -815,7 +887,7 @@ class NDCube(NDCubeBase):
         Default is None.
 
     extra_coords : iterable of `tuple`, each with three entries
-        (`str`, `int`, `astropy.units.quantity` or array-like)
+        (`str`, `int`, `astropy.units.Quantity` or array-like)
         Gives the name, axis of data, and values of coordinates of a data axis not
         included in the WCS object.
 
@@ -848,13 +920,13 @@ class NDCube(NDCubeBase):
     def _as_mpl_axes(self):
         if hasattr(self.plotter, "_as_mpl_axes"):
             return self.plotter._as_mpl_axes()
-        else:
-            warnings.warn(f"The current plotter {self.plotter} does not have a '_as_mpl_axes' method. "
-                          "The default MatplotlibPlotter._as_mpl_axes method will be used instead.",
-                          UserWarning)
+        warn_user(f"The current plotter {self.plotter} does not have a '_as_mpl_axes' method. "
+                    "The default MatplotlibPlotter._as_mpl_axes method will be used instead.")
 
-            plotter = MatplotlibPlotter(self)
-            return plotter._as_mpl_axes()
+        from ndcube.visualization.mpl_plotter import MatplotlibPlotter
+
+        plotter = MatplotlibPlotter(self)
+        return plotter._as_mpl_axes()
 
     def plot(self, *args, **kwargs):
         """
@@ -872,15 +944,17 @@ class NDCube(NDCubeBase):
 
         return self.plotter.plot(*args, **kwargs)
 
-    def _new_instance_from_op(self, new_data, new_unit, new_uncertainty):
-        # This implicitly assumes that the arithmetic operation does not alter
-        # the WCS, mask, or metadata.
-        new_cube = type(self)(new_data,
-                              unit=new_unit,
-                              wcs=self.wcs,
-                              mask=deepcopy(self.mask),
-                              meta=deepcopy(self.meta),
-                              uncertainty=new_uncertainty)
+    def _new_instance(self, **kwargs):
+        keys = ('unit', 'wcs', 'mask', 'meta', 'uncertainty', 'psf')
+        new_kwargs = {k: deepcopy(getattr(self, k, None)) for k in keys}
+        # To support old versions of astropy, we need to make sure
+        # we only pass in the parameters that are valid for the NDData
+        params = list(inspect.signature(astropy.nddata.NDData).parameters)
+        full_kwargs = {x: new_kwargs.pop(x) for x in params & new_kwargs.keys()}
+        # We Explicitly DO NOT deepcopy any data
+        full_kwargs['data'] = self.data
+        full_kwargs.update(kwargs)
+        new_cube = type(self)(**full_kwargs)
         if self.extra_coords is not None:
             new_cube._extra_coords = deepcopy(self.extra_coords)
         if self.global_coords is not None:
@@ -888,8 +962,7 @@ class NDCube(NDCubeBase):
         return new_cube
 
     def __neg__(self):
-        return self._new_instance_from_op(-self.data, deepcopy(self.unit),
-                                          deepcopy(self.uncertainty))
+        return self._new_instance(data=-self.data)
 
     def __add__(self, value):
         if hasattr(value, 'unit'):
@@ -908,7 +981,7 @@ class NDCube(NDCubeBase):
             raise TypeError("Cannot add a unitless object to an NDCube with a unit.")
         else:
             new_data = self.data + value
-        return self._new_instance_from_op(new_data, deepcopy(self.unit), deepcopy(self.uncertainty))
+        return self._new_instance(data=new_data)
 
     def __radd__(self, value):
         return self.__add__(value)
@@ -936,8 +1009,7 @@ class NDCube(NDCubeBase):
         new_data = self.data * value
         new_uncertainty = (type(self.uncertainty)(self.uncertainty.array * value)
                            if self.uncertainty is not None else None)
-        new_cube = self._new_instance_from_op(new_data, new_unit, new_uncertainty)
-        return new_cube
+        return self._new_instance(data=new_data, unit=new_unit, uncertainty=new_uncertainty)
 
     def __rmul__(self, value):
         return self.__mul__(value)
@@ -945,6 +1017,30 @@ class NDCube(NDCubeBase):
     def __truediv__(self, value):
         return self.__mul__(1/value)
 
+    def __rtruediv__(self, value):
+        return self.__pow__(-1).__mul__(value)
+
+    def __pow__(self, value):
+        new_data = self.data ** value
+        new_unit = self.unit if self.unit is None else self.unit ** value
+        new_uncertainty = self.uncertainty
+
+        if self.uncertainty is not None:
+            try:
+                new_uncertainty = new_uncertainty.propagate(np.power, self, self.data ** value, correlation=1)
+            except ValueError as e:
+                if "unsupported operation" in e.args[0]:
+                    new_uncertainty = None
+                    warn_user(f"{type(self.uncertainty)} does not support propagation of uncertainties for power. "
+                              f"Setting uncertainties to None.")
+                elif "does not support uncertainty propagation" in e.args[0]:
+                    new_uncertainty = None
+                    warn_user(f"{e.args[0]} Setting uncertainties to None.")
+                else:
+                    raise e
+
+        return self._new_instance(data=new_data, unit=new_unit, uncertainty=new_uncertainty)
+
     def to(self, new_unit, **kwargs):
         """Convert instance to another unit.
 
@@ -979,11 +1075,14 @@ class NDCube(NDCubeBase):
 
         Parameters
         ----------
-        bin_shape : array-like
+        bin_shape : array-like, `astropy.units.Quantity`
             The number of pixels in a bin in each dimension.
             Must be the same length as number of dimensions in data.
             Each element must be in int. If they are not they will be rounded
-            to the nearest int.
+            to the nearest int. If provided as a `~astropy.units.Quantity` the
+            units have to be convertible to pixels.
+            The sentinel value ``-1`` can be passed for a dimension which means
+            that no rebinning will occur along that dimension.
         operation : function
             Function applied to the data to derive values of the bins.
             Default is `numpy.mean`
@@ -1082,37 +1181,32 @@ class NDCube(NDCubeBase):
         """
         # Sanitize input.
         new_unit = new_unit or self.unit
+        if isinstance(bin_shape, u.Quantity):
+            bin_shape = bin_shape.to_value(u.pixel)
         # Make sure the input bin dimensions are integers.
         bin_shape = np.rint(bin_shape).astype(int)
-        offsets = (bin_shape - 1) / 2
-        if all(bin_shape == 1):
+        if np.all(bin_shape == 1):
             return self
         # Ensure bin_size has right number of entries and each entry is an
         # integer fraction of the array shape in each dimension.
-        data_shape = self.dimensions.value.astype(int)
+        data_shape = self.shape
         naxes = len(data_shape)
         if len(bin_shape) != naxes:
             raise ValueError("bin_shape must have an entry for each array axis.")
+        bin_shape[bin_shape == -1] = np.array(data_shape)[bin_shape == -1]
+        if (bin_shape < 0).any():
+            raise ValueError("bin_shape should not be less than -1.")
         if (np.mod(data_shape, bin_shape) != 0).any():
             raise ValueError(
                 "bin shape must be an integer fraction of the data shape in each dimension. "
-                f"data shape: {data_shape};  bin shape: {bin_shape}")
+                f"data shape: {data_shape};  bin shape: {bin_shape}"
+            )
 
         # Reshape array so odd dimensions represent pixels to be binned
         # then apply function over those axes.
-        m = None if (self.mask is None or self.mask is False or operation_ignores_mask) else self.mask
-        data = self.data
-        if m is not None:
-            for array_type, masked_type in ARRAY_MASK_MAP.items():
-                if isinstance(self.data, array_type):
-                    break
-            else:
-                masked_type = np.ma.masked_array
-                warn.warning("data and mask arrays of different or unrecognized types. "
-                             "Casting them into a numpy masked array.")
-            data = masked_type(self.data, m)
-
-        reshape = np.empty(data_shape.size + bin_shape.size, dtype=int)
+        data, sanitized_mask = _create_masked_array_for_rebinning(self.data, self.mask,
+                                                                  operation_ignores_mask)
+        reshape = np.empty(len(data_shape) + len(bin_shape), dtype=int)
         new_shape = (data_shape / bin_shape).astype(int)
         reshape[0::2] = new_shape
         reshape[1::2] = bin_shape
@@ -1134,16 +1228,17 @@ class NDCube(NDCubeBase):
         new_uncertainty = None
         if propagate_uncertainties:
             if self.uncertainty is None:
-                warnings.warn("Uncertainties cannot be propagated as there are no uncertainties, "
-                              "i.e. self.uncertainty is None.")
+                warn_user("Uncertainties cannot be propagated as there are no uncertainties, "
+                              "i.e., the `uncertainty` keyword was never set on creation of this NDCube.")
             elif isinstance(self.uncertainty, astropy.nddata.UnknownUncertainty):
-                warnings.warn("self.uncertainty is of type UnknownUncertainty which does not "
-                              "support uncertainty propagation.")
+                warn_user("The uncertainty on this NDCube has no known way to propagate forward and so will be dropped."
+                              "To create an uncertainty that can propagate, please see "
+                              "https://docs.astropy.org/en/stable/uncertainty/index.html")
             elif (not operation_ignores_mask
                   and (self.mask is True or (self.mask is not None
                                              and not isinstance(self.mask, bool)
                                              and self.mask.all()))):
-                warnings.warn("Uncertainties cannot be propagated as all values are masked and "
+                warn_user("Uncertainties cannot be propagated as all values are masked and "
                               "operation_ignores_mask is False.")
             else:
                 if propagate_uncertainties is True:
@@ -1155,7 +1250,7 @@ class NDCube(NDCubeBase):
                 # in each bin can be iterated (all bins being treated in parallel) and
                 # their uncertainties propagated.
                 bin_size = bin_shape.prod()
-                flat_shape = [bin_size] + list(new_shape)
+                flat_shape = [bin_size, *list(new_shape)]
                 dummy_axes = tuple(range(1, len(reshape), 2))
                 flat_data = np.moveaxis(reshaped_data, dummy_axes, tuple(range(naxes)))
                 flat_data = flat_data.reshape(flat_shape)
@@ -1163,7 +1258,7 @@ class NDCube(NDCubeBase):
                 flat_uncertainty = np.moveaxis(reshaped_uncertainty, dummy_axes, tuple(range(naxes)))
                 flat_uncertainty = flat_uncertainty.reshape(flat_shape)
                 flat_uncertainty = type(self.uncertainty)(flat_uncertainty)
-                if m is not None:
+                if sanitized_mask is not None:
                     reshaped_mask = self.mask.reshape(tuple(reshape))
                     flat_mask = np.moveaxis(reshaped_mask, dummy_axes, tuple(range(naxes)))
                     flat_mask = flat_mask.reshape(flat_shape)
@@ -1178,15 +1273,71 @@ class NDCube(NDCubeBase):
         # Resample WCS
         new_wcs = ResampledLowLevelWCS(self.wcs.low_level_wcs, bin_shape[::-1])
 
+        # If meta is axis-aware, drop axis-awareness for metadata associated with rebinned axes.
+        if hasattr(self.meta, "__ndcube_can_rebin__") and self.meta.__ndcube_can_rebin__:
+            new_meta = self.meta.rebin(bin_shape)
+        else:
+            new_meta = deepcopy(self.meta)
+
         # Reform NDCube.
-        new_cube = type(self)(data=new_data, wcs=new_wcs, uncertainty=new_uncertainty, mask=new_mask,
-                              meta=self.meta, unit=new_unit)
+        new_cube = type(self)(
+            data=new_data,
+            wcs=new_wcs,
+            uncertainty=new_uncertainty,
+            mask=new_mask,
+            meta=new_meta,
+            unit=new_unit
+        )
         new_cube._global_coords = self._global_coords
         # Reconstitute extra coords
         if not self.extra_coords.is_empty:
-            new_array_grids = [None if bin_shape[i] == 1 else
-                               np.arange(offsets[i], data_shape[i] + offsets[i], bin_shape[i])
-                               for i in range(naxes)]
             new_cube._extra_coords = self.extra_coords.resample(bin_shape, ndcube=new_cube)
 
         return new_cube
+
+    def squeeze(self, axis=None):
+        """
+        Removes all axes with a length of 1.
+
+        Parameters
+        ----------
+        axis: array-like, optional
+            Specifies specific axes to be removed. If one of those axes
+            has length larger than 1, an error is thrown. If not specified,
+            all axes of length 1 get removed.
+
+        Returns
+        -------
+        `~ndcube.NDCube`
+            A new NDCube instance with the removed axes.
+        """
+        item = np.full(self.data.ndim, slice(None))
+        shape = np.asarray(self.data.shape)
+        if axis is None:
+            item[shape == 1] = 0
+        else:
+            # For simplicity’s sake, if the axis is scalar make it a tuple.
+            if isinstance(axis, numbers.Integral):
+                axis = (axis,)
+            axis = np.asarray(axis)
+            if not (shape[axis] == 1).all():
+                raise ValueError("Cannot select any axis to squeeze out, as none of them has size equal to one.")
+            item[axis] = 0
+        # Scalar NDCubes are not supported, so we raise error as the operation would cause all the axes to be squeezed.
+        if (item == 0).all():
+            raise ValueError("All axes are of length 1, therefore we will not squeeze NDCube to become a scalar. "
+                             "Use `axis=` keyword to specify a subset of axes to squeeze.")
+        return self[tuple(item)]
+
+
+def _create_masked_array_for_rebinning(data, mask, operation_ignores_mask):
+    m = None if (mask is None or mask is False or operation_ignores_mask) else mask
+    if m is None:
+        return data, m
+    for array_type, masked_type in ARRAY_MASK_MAP.items():
+        if isinstance(data, array_type):
+            break
+    else:
+        masked_type = np.ma.masked_array
+        warn_user("data and mask arrays of different or unrecognized types. Casting them into a numpy masked array.")
+    return masked_type(data, m), m
